#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Drawing.Imaging;
using System.Linq;
using System.Windows.Forms;

namespace HeuristicLab.Visualization {
  public partial class ChartControl : UserControl {
    public static readonly Cursor HandCursor = new Cursor(typeof(ChartControl), "Resources.Hand.cur");
    public static readonly Cursor ZoomOutCursor = new Cursor(typeof(ChartControl), "Resources.ZoomOut.cur");
    public static readonly Cursor ZoomInCursor = new Cursor(typeof(ChartControl), "Resources.ZoomIn.cur");

    protected Bitmap Picture;
    protected bool RenderingRequired;
    protected Point ButtonDownPoint;
    protected Point PreviousLocation;
    protected int MouseClickCount;

    protected bool SuppressEvents { get; set; }
    private bool SuppressRender { get; set; }

    private IChart chart;
    public IChart Chart {
      get { return chart; }
      protected set {
        if (chart != null) DeregisterChartEvents(chart);
        chart = value;
        chart.Enabled = Enabled;
        pictureBox.Enabled = Enabled;
        if (chart != null) RegisterChartEvents(chart);
      }
    }

    private ChartMode mode;
    public ChartMode Mode {
      get { return mode; }
      set {
        if (mode == value) return;
        mode = value;
        SetButtonCheckedSate();
        SetPictureBoxCursor(GetPictureBoxCursor());
        OnModeChanged();
        if (chart != null) {
          GenerateImage();
        }
      }
    }

    public bool ScaleOnResize { get; set; }

    public bool IsRenderingPaused { get { return SuppressRender; } }

    public ChartControl() : this(null) { }
    public ChartControl(IChart defaultChart) {
      InitializeComponent();
      ScaleOnResize = true;
      Mode = ChartMode.Select;
      Chart = defaultChart ?? new Chart(0, 0, Width, Height);
    }

    public void SuspendRendering() {
      SuppressRender = true;
    }

    public void ResumeRendering() {
      SuppressRender = false;
      if (Chart != null) GenerateImage();
    }

    protected virtual void SetButtonCheckedSate() {
      selectButton.Checked = false;
      moveButton.Checked = false;
      zoomInButton.Checked = false;
      zoomOutButton.Checked = false;
      if (mode == ChartMode.Select) selectButton.Checked = true;
      else if (mode == ChartMode.Move) moveButton.Checked = true;
      else if (mode == ChartMode.ZoomIn) zoomInButton.Checked = true;
      else if (mode == ChartMode.ZoomOut) zoomOutButton.Checked = true;
    }

    protected virtual void SetPictureBoxCursor(Cursor cursor) {
      if (pictureBox.Cursor != cursor) pictureBox.Cursor = cursor;
    }

    protected virtual Cursor GetPictureBoxCursor() {
      if (Mode == ChartMode.Select)
        return DefaultCursor;
      if (Mode == ChartMode.Move)
        return HandCursor;
      if (Mode == ChartMode.ZoomIn)
        return ZoomInCursor;
      if (Mode == ChartMode.ZoomOut)
        return ZoomOutCursor;
      return null;
    }

    protected virtual void RegisterChartEvents(IChart chart) {
      chart.RedrawRequired += ChartOnUpdateRequired;
    }

    protected virtual void DeregisterChartEvents(IChart chart) {
      chart.RedrawRequired -= ChartOnUpdateRequired;
    }

    protected virtual void ChartOnUpdateRequired(object sender, EventArgs e) {
      if (SuppressRender) return;
      GenerateImage();
    }

    #region PictureBox Events
    protected virtual void PictureBoxOnSizeChanged(object sender, EventArgs e) {
      if (Chart == null) return;
      if (ScaleOnResize) {
        if ((pictureBox.Width > 0) && (pictureBox.Height > 0) && (Chart != null)) {
          var point = Chart.TransformPixelToWorld(new Point(pictureBox.Width, Chart.SizeInPixels.Height - pictureBox.Height));
          var diff = Chart.UpperRight - point;
          Chart.SetPosition(new PointD(Chart.LowerLeft.X + (diff.DX / 2.0), Chart.LowerLeft.Y + (diff.DY / 2.0)),
            new PointD(Chart.UpperRight.X - (diff.DX / 2.0), Chart.UpperRight.Y - (diff.DY / 2.0)));
        }
      }
      GenerateImage();
    }

    protected virtual void PictureBoxOnVisibleChanged(object sender, EventArgs e) {
      if (pictureBox.Visible && RenderingRequired) GenerateImage();
    }

    protected virtual void PictureBoxOnMouseClick(object sender, MouseEventArgs e) {

    }

    protected virtual void PictureBoxOnMouseDoubleClick(object sender, MouseEventArgs e) {

    }

    protected virtual void OnKeyDown(object sender, KeyEventArgs e) {
      if (Mode == ChartMode.Select) {
        if (e.KeyCode == Keys.Delete) {
          try {
            SuspendRendering();
            foreach (var primitive in Chart.Group.SelectedPrimitives)
              Chart.Group.Remove(primitive);
          } finally { ResumeRendering(); }
        }
      }
    }

    protected virtual void OnKeyUp(object sender, KeyEventArgs e) {

    }

    protected virtual void PictureBoxOnMouseWheel(object sender, MouseEventArgs e) {

    }

    protected virtual void PictureBoxOnMouseDown(object sender, MouseEventArgs e) {
      if (Chart == null) return;
      MouseClickCount = e.Clicks;
      ButtonDownPoint = e.Location;
      if (Mode != ChartMode.Select) return;

      try {
        SuspendRendering();
        var worldLocation = Chart.TransformPixelToWorld(e.Location);
        foreach (var sp in Chart.Group.SelectedPrimitives.Where(x => !x.ContainsPoint(worldLocation)))
          sp.Selected = false;
        var p = Chart.GetPrimitive(e.Location);
        if (p != null) p.Selected = true;
      } finally { ResumeRendering(); }
    }

    protected virtual void PictureBoxOnMouseUp(object sender, MouseEventArgs e) {
      if (Chart == null) return;
      if (e.Button == MouseButtons.Left) {
        if (Mode == ChartMode.ZoomIn) {
          //pictureBox.Refresh();
          var lowerLeft = new Point(Math.Min(e.X, ButtonDownPoint.X), Math.Max(e.Y, ButtonDownPoint.Y));
          var upperRight = new Point(Math.Max(e.X, ButtonDownPoint.X), Math.Min(e.Y, ButtonDownPoint.Y));
          if ((lowerLeft.X != upperRight.X) && (lowerLeft.Y != upperRight.Y))
            Chart.ZoomIn(lowerLeft, upperRight);
          else Chart.ZoomIn(ButtonDownPoint);
        } else if (Mode == ChartMode.ZoomOut) {
          Chart.ZoomOut();
        }
      } else if (e.Button == MouseButtons.Right) {
        propertiesToolStripMenuItem.Enabled = Chart.Group.SelectedPrimitives.Count() == 1;
        oneLayerUpToolStripMenuItem.Enabled = Chart.Group.SelectedPrimitives.Count() == 1;
        oneLayerDownToolStripMenuItem.Enabled = Chart.Group.SelectedPrimitives.Count() == 1;
        intoForegroundToolStripMenuItem.Enabled = Chart.Group.SelectedPrimitives.Count() == 1;
        intoBackgroundToolStripMenuItem.Enabled = Chart.Group.SelectedPrimitives.Count() == 1;
      } else if (e.Button == MouseButtons.Middle) {
        if (Mode == ChartMode.Select) {
          try {
            SuspendRendering();
            foreach (var p in Chart.Group.SelectedPrimitives)
              p.Selected = false;
          } finally { ResumeRendering(); }
        } else if (Mode == ChartMode.Move || Mode == ChartMode.ZoomIn || Mode == ChartMode.ZoomOut) {
          Chart.Unzoom();
        }
      }
    }

    protected virtual void PictureBoxOnMouseMove(object sender, MouseEventArgs e) {
      if (InvokeRequired) {
        Invoke((Action<Object, MouseEventArgs>)PictureBoxOnMouseMove, sender, e);
        return;
      }
      var toolTipText = Chart.GetToolTipText(e.Location);
      if (toolTip.GetToolTip(pictureBox) != toolTipText) toolTip.SetToolTip(pictureBox, toolTipText);

      Cursor cursor;
      if (Mode == ChartMode.Select) {
        cursor = Chart.GetCursor(e.Location) ?? GetPictureBoxCursor();
      } else cursor = GetPictureBoxCursor();
      SetPictureBoxCursor(cursor);

      try {
        if (e.Button == MouseButtons.None) return;

        if (Mode == ChartMode.Select && e.Button == MouseButtons.Left) {
          var previousWorldLocation = Chart.TransformPixelToWorld(PreviousLocation);
          var worldLocation = Chart.TransformPixelToWorld(e.Location);
          var offset = worldLocation - previousWorldLocation;
          try {
            SuspendRendering();
            foreach (var primitive in Chart.Group.SelectedPrimitives.Where(p => p.ContainsPoint(previousWorldLocation)))
              primitive.Move(previousWorldLocation, offset);
          } finally { ResumeRendering(); }
        } else if (Mode == ChartMode.Move) {
          var previousWorldLocation = Chart.TransformPixelToWorld(PreviousLocation);
          var worldLocation = Chart.TransformPixelToWorld(e.Location);
          var offset = worldLocation - previousWorldLocation;
          Chart.Move(new Offset(-1 * offset.DX, -1 * offset.DY));
        } else if (Mode == ChartMode.ZoomIn && e.Button == MouseButtons.Left) {
          pictureBox.Refresh();
          using (var graphics = pictureBox.CreateGraphics()) {
            using (var pen = new Pen(Color.Gray) { DashStyle = DashStyle.Dash }) {
              graphics.DrawRectangle(pen,
                Math.Min(e.X, ButtonDownPoint.X),
                Math.Min(e.Y, ButtonDownPoint.Y),
                Math.Abs(e.X - ButtonDownPoint.X),
                Math.Abs(e.Y - ButtonDownPoint.Y));
            }
          }
        }
      } finally {
        PreviousLocation = e.Location;
        pictureBox.Update();
      }
    }

    protected virtual void PictureBoxOnMouseEnter(object sender, EventArgs e) {
      if (InvokeRequired) {
        Invoke((Action<Object, MouseEventArgs>)PictureBoxOnMouseEnter, sender, e);
        return;
      }
      if (!Focused) this.pictureBox.Focus();
    }

    protected virtual void PictureBoxOnMouseLeave(object sender, EventArgs e) {

    }

    protected virtual void PictureBoxOnClick(object sender, EventArgs e) {
      if (InvokeRequired) {
        Invoke((Action<Object, MouseEventArgs>)PictureBoxOnClick, sender, e);
        return;
      }
      if (!Focused) this.pictureBox.Focus();
    }
    #endregion

    protected virtual void PictureBoxContextMenuStripOnOpening(object sender, CancelEventArgs e) { }

    protected virtual void oneLayerUpToolStripMenuItem_Click(object sender, EventArgs e) {
      if (Chart == null) return;
      if (Chart.Group.SelectedPrimitives.Count() == 1) {
        Chart.OneLayerUp(Chart.Group.SelectedPrimitives.First());
      }
    }
    protected virtual void oneLayerDownToolStripMenuItem_Click(object sender, EventArgs e) {
      if (Chart == null) return;
      if (Chart.Group.SelectedPrimitives.Count() == 1) {
        Chart.OneLayerDown(Chart.Group.SelectedPrimitives.First());
      }
    }
    protected virtual void intoForegroundToolStripMenuItem_Click(object sender, EventArgs e) {
      if (Chart == null) return;
      if (Chart.Group.SelectedPrimitives.Count() == 1) {
        Chart.IntoForeground(Chart.Group.SelectedPrimitives.First());
      }
    }
    protected virtual void intoBackgroundToolStripMenuItem_Click(object sender, EventArgs e) {
      if (Chart == null) return;
      if (Chart.Group.SelectedPrimitives.Count() == 1) {
        Chart.IntoBackground(Chart.Group.SelectedPrimitives.First());
      }
    }
    protected virtual void propertiesToolStripMenuItem_Click(object sender, EventArgs e) {
      if (Chart == null) return;
      if (Chart.Group.SelectedPrimitives.Count() == 1) {
        using (var dialog = new PropertiesDialog(Chart.Group.SelectedPrimitives.First())) {
          dialog.ShowDialog(this);
        }
      }
    }
    protected virtual void saveImageToolStripMenuItem_Click(object sender, EventArgs e) {
      if (Picture == null) return;

      if (saveFileDialog.ShowDialog() == DialogResult.OK) {
        var format = ImageFormat.Bmp;
        var filename = saveFileDialog.FileName.ToLower();
        if (filename.EndsWith("emf")) {
          using (var g1 = Graphics.FromImage(Picture)) {
            var rectangle = new System.Drawing.Rectangle(0, 0, Picture.Width, Picture.Height);
            using (var metafile = new Metafile(filename, g1.GetHdc(), rectangle, MetafileFrameUnit.Pixel, EmfType.EmfPlusDual))
            using (var g2 = Graphics.FromImage(metafile))
              Chart.Render(g2, pictureBox.Width, pictureBox.Height);
          }
        } else {
          using (var graphics = Graphics.FromImage(Picture))
            Chart.Render(graphics, Picture.Width, Picture.Height);

          if (filename.EndsWith("jpg")) {
            format = ImageFormat.Jpeg;
          } else if (filename.EndsWith("gif")) {
            format = ImageFormat.Gif;
          } else if (filename.EndsWith("png")) {
            format = ImageFormat.Png;
          } else if (filename.EndsWith("tif")) {
            format = ImageFormat.Tiff;
          }

          Picture.Save(saveFileDialog.FileName, format);
        }
      }
    }

    protected virtual void GenerateImage() {
      if (InvokeRequired) {
        Invoke((Action)GenerateImage);
        return;
      }
      if (!Visible) {
        RenderingRequired = true;
      } else {
        if ((pictureBox.Width == 0) || (pictureBox.Height == 0)) {
          pictureBox.Image = pictureBox.ErrorImage;
          if (Picture != null) Picture.Dispose();
          Picture = null;
        } else {
          if (Picture != null) Picture.Dispose();
          Picture = new Bitmap(pictureBox.Width, pictureBox.Height);
          using (var graphics = Graphics.FromImage(Picture)) {
            graphics.SetClip(new System.Drawing.Rectangle(0, 0, pictureBox.Width, pictureBox.Height));
            Chart.Render(graphics, pictureBox.Width, pictureBox.Height);
          }
        }
        pictureBox.Image = Picture;
        RenderingRequired = false;
      }
    }

    public event EventHandler ModeChanged;
    protected void OnModeChanged() {
      var handler = ModeChanged;
      if (handler != null) handler(this, EventArgs.Empty);
    }

    protected virtual void SelectButtonOnCheckedChanged(object sender, EventArgs e) {
      if (selectButton.Checked) Mode = ChartMode.Select;
    }

    protected virtual void MoveButtonOnCheckedChanged(object sender, EventArgs e) {
      if (moveButton.Checked) Mode = ChartMode.Move;
    }

    protected virtual void ZoomInButtonOnCheckedChanged(object sender, EventArgs e) {
      if (zoomInButton.Checked) Mode = ChartMode.ZoomIn;
    }

    protected virtual void ZoomOutButtonOnCheckedChanged(object sender, EventArgs e) {
      if (zoomOutButton.Checked) Mode = ChartMode.ZoomOut;
    }

    private void ChartControl_Load(object sender, EventArgs e) {
      if (!SuppressRender) GenerateImage();
    }
  }
}
